#include <stdlib.h>
#include <stdarg.h>
#include <windows.h>
#include "fish.h"
#include "compress.hpp"
#include "alloc.h"

// The STUB CRC32 should be as defined. Each time the stub is changed this
// number should be changed too.
#define STUB_CRC       0xE969A735

// A debugging macro. Does nothing in the final version
#define HEAPCHECK

ulong scrc32;    // the real CRC32 of the stub

//-----------------------------------------------------------------------

#ifdef DEBUG
int debug = 1;

void deb(const char *format, ...)
{
  static FILE *fp = NULL;
  if ( fp == NULL ) fp = fopen("zzz", "a");
  if ( fp != NULL )
  {
    va_list va;
    va_start(va, format);
    vfprintf(fp, format, va);
    va_end(va);
    fflush(fp);
  }
}

void hex(const char *msg, uchar *iv, int size)
{
  deb(msg);
  for ( int i=0; i < size; i++ )
    deb(" %02X", iv[i]);
  deb("\n");
}

void DebugIO(char *s) { deb("%s", s); }
#else
inline void deb(const char *,...) {}
inline void hex(const char *,uchar *, int) {}
#endif

//-----------------------------------------------------------------------
// Produce random initialization vector
void produce_iv(uchar *iv)
{
  for ( int i=0; i < KEYSIZE/4; i++ )
    ((ulong*)iv)[i] = Rand();
}

//-----------------------------------------------------------------------
// Encrypt/decrypt a buffer using TWOFISH
//    encrypt - true: encrypt, false:decrypt
//    bytes   - pointer to input/output buffer
//    len     - length of buffer in bytes
//    phrase  - the passphrase to derive the encryption key
//    callback- a function to display the progress of the operation
int cipher(bool encrypt,uchar *iv, uchar *bytes, int len, const char *phrase,
                                        bool callback(int))
{
  deb("\nchiper %d bytes=%p len=%d phrase='%s'\n", encrypt, bytes, len, phrase);
  if ( len == 64 ) hex("in:", bytes, len);
  hex("iv:", iv, KEYSIZE);
  cipherInstance c;
  keyInstance k;

  char a[MAX_KEY_SIZE];
  int code = cipherInit(&c, MODE_CBC, b2a(iv,a));
  deb("cipherinit %s\n", a);
  if ( code != TRUE ) return code;

  uchar hash[KEYSIZE];
  hashpass(phrase, strlen(phrase), hash);
  hex("ha:", hash, KEYSIZE);
  code = makeKey(&k, encrypt ? DIR_ENCRYPT : DIR_DECRYPT, KEYBITS, b2a(hash,a));
  deb("makeKey %s\n", a);
  if ( code != TRUE ) return code;

  while ( len > 0 )
  {
    const int CHUNK = 0x10000000;
    int chunk = (len < CHUNK) ? len : CHUNK;
    // round the length up to next block size
    int bits = len * 8;
    bits += BLOCK_SIZE - 1;
    bits &= ~(BLOCK_SIZE - 1);

    // encrypt block
    code = (encrypt ? blockEncrypt : blockDecrypt)
                        (&c, &k, bytes, bits, bytes, callback);  // twofish2.cpp
    if ( code != bits ) return code;

    if ( len == 64 ) hex(encrypt ? "en:" : "de:", bytes, len);
    len -= chunk;
    bytes += chunk;
  }

  memset(&c, 0, sizeof(c));
  memset(&k, 0, sizeof(k));
  return TRUE;
}

//-----------------------------------------------------------------------
// The rest of the file
//     - deals with the compression
//     - prepares an image to encrypt
//     - prepends a stub to it
//     - writes out a resulting file

// pointer callback function to display the progress
static bool (*_callback)(int type, int size, int fullsize);

// input file
static FILE *fp;

// The whole payload of the resulting file is stored in the memory in a temproary
// buffer. "memory" points to this payload.
// The payload looks like this:
//      header_t
//      message
//x     header2_t
//x     fileh_t     -+
//x     file name    |   many times
//x     ciphertext  -+
//      "ENCRYPTED_FILE"
static char *memory = NULL;

// The payload starts with a header_t which is followed by an optional message
// attached to the archive.
static header_t *h;

// After the message there is another header which specifies the number of files
static header2_t *h2;

// offset to the current file header
static int fileh_off;

// get pointer to the current file header
inline fileh_t *fh(void) { return (fileh_t*)(memory + fileh_off); }

static int memory_size;        // total size of allocated memory
static int header_size;        // h + msg + h2
static int fileh_size;         // h + msg + h2
static int zipped_size;        // memory used so far
static int read_size;          // number of bytes read from the current file

// Callback for ZIP: read a block
static unsigned file_read(void *buf, unsigned size)
{
  read_size += size;
  if ( !_callback(CB_ZIP, read_size, fh()->usize) )
  {
    read_size = -1;
    return 0;
  }
  return fread(buf,1,size,fp);
}

// Make sure there is enough memory to store "delta" bytes
static void makemem(ulong delta)
{
  int total_size = header_size + zipped_size + delta;
  if ( memory_size < total_size )
  {
    if ( fileh_off ) total_size += fh()->usize/2;
    char *m2 = (char*)realloc(memory, total_size+KEYSIZE);
    if ( m2 == NULL ) nomem("zip add");
    memory = m2;
    h = (header_t *)memory;
    h2 = h->h2();
    memory_size = total_size;
  }
}

// Callback for ZIP: write a block
static unsigned mem_write(void *buf, unsigned size)
{
  makemem(size);
  memcpy(memory+header_size+zipped_size, buf, size);
  zipped_size += size;
  return size;
//  return fwrite(buf,1,size,out);
}

// Allocate a new file header
static void alloc_fileh(void)
{
  fileh_off = 0;
  makemem(fileh_size);
  fileh_off = header_size+zipped_size;
  zipped_size += fileh_size;
}

//-----------------------------------------------------------------------
// A thunk to be called from the encryption routines
static bool enc_callback(int bits)
{
  return _callback(CB_FISH, bits/8, h->size);
}

//-----------------------------------------------------------------------
// Copy the executable stub to the output file.
// The current version uses the stub bound into aCrypt as a resource.
// The input parameter "stub" is not used.
static int copy_stub(FILE *ofp, const char *stub)
{
  _callback(CB_STUB, 0, 3);
#if 0  // file version      ------- NOT USED NOW
  fp = fopen(stub, "rb");
  if ( fp == NULL ) return ERR_STUB;

  size_t ssize = filelength(fileno(fp));
  char *sbuf = new char[ssize];
  if ( sbuf == NULL ) return ERR_NOMEM;
  if ( fread(sbuf, 1, ssize, fp) != ssize )
  {
    delete sbuf;
    return ERR_READ;
  }
  crc32(NULL, 0);
  scrc32 = crc32(sbuf, ssize);
  h->stub_crc32 = scrc32;
  if ( scrc32 != STUB_CRC )
  {
#if 0
    fclose(ofp);
    unlink(outfile);
    return ERR_BADSTUB;
#else
    extern const char a_badstub[];
    warning(a_badstub, scrc32);
#endif
  }
  _callback(CB_STUB, 1, 3);
  if ( fwrite(sbuf, 1, ssize, ofp) != ssize )
  {
    delete sbuf;
    return ERR_WRITE;
  }
  _callback(CB_STUB, 2, 3);
  delete sbuf;
  fclose(fp);
#else        // resource version   -------- THIS VERSION IS USED NOW
  static char *sbuf = NULL;
  static size_t ssize;
  // If the stub was not retrieved from the resources, do it now
  if ( sbuf == NULL )
  {
    // The stub has resource number 777. Why not? :)
    HRSRC hr = FindResource(NULL, MAKEINTRESOURCE(777), RT_RCDATA);
    sbuf = (char *)LockResource(LoadResource(NULL, hr));
    ssize = SizeofResource(NULL, hr);
  }
  crc32(NULL, 0);
  scrc32 = crc32(sbuf, ssize);
  h->stub_crc32 = scrc32;
  // Complain if the stub CRC32 is not equal to the expected value.
  // We are not going to produce bad executables.
  if ( scrc32 != STUB_CRC )
  {
    extern const char a_badstub[];
    warning(a_badstub, scrc32);
  }
  _callback(CB_STUB, 1, 3);
  // Write the stub to the output file
  if ( fwrite(sbuf, 1, ssize, ofp) != ssize ) return ERR_WRITE;
  _callback(CB_STUB, 2, 3);
#endif
  _callback(CB_STUB, 3, 3);
  return ERR_OK;
}

//-----------------------------------------------------------------------
// All the input files are read and compressed.
// Encrypt them and create the output file.
// Returns: error code ERR_...
int finish_zip(const char *outfile, const char *phrase,
                        const char *stub, bool callback(int,int,int))
{
  _callback = callback;
  h->size     = sizeof(header2_t) + zipped_size;
  h->size    += KEYSIZE-1;
  h->size    &= ~(KEYSIZE-1);
//  warning("zipped=%ld h->size=%ld h=%x h2=%x", zipped_size, h->size, h, h2);

  // encrypt the payload
  int code = cipher(true, h->iv, (uchar*)h2, h->size, phrase, enc_callback);
  if ( code == BAD_CANCEL ) return ERR_CANCEL;

  // open the output file
  FILE *ofp = fopen(outfile, "wb");
  if ( ofp == NULL ) return ERR_CRFILE;

  // copy the stub
  code = copy_stub(ofp, stub);
  if ( code != ERR_OK ) return code;

  // CRC32 protect the payload
  crc32(NULL, 0);                                  // initialize crc
  h->crc32 = crc32(h->crc_start(), h->crc_len());  // and calc crc
//  warning("crc32=%08lX size=%d", h->crc32, h->size);

  // write out the payload
  unsigned int osize = sizeof(header_t) + h->msglen + h->size;
  if ( fwrite(h, 1, osize, ofp) != osize ) return ERR_WRITE;

  // write out the singature saying that the file is encrypted
  if ( fwrite(ENCRYPTED_SIGN, 1, ENCRYPTED_SIGN_SIZE, ofp) != ENCRYPTED_SIGN_SIZE )
    return ERR_WRITE;
  callback(CB_END, 1, 1);
  fclose(ofp);

  // cleanup and return
  free(memory);
  memory = NULL;
  h  = NULL;
  h2 = NULL;

  return ERR_OK;
}

//-----------------------------------------------------------------------
// Read and compress one file
// Returns: error code ERR_...
int zip_one_file(const char *file, const char *filename, bool callback(int,int,int))
{
  HEAPCHECK;
  _callback = callback;
  fp = fopen(file, "rb");
  if ( fp == NULL ) return ERR_NOFILE;

  // create a file header
  int namelen = strlen(filename) + 1;
  fileh_size  = sizeof(fileh_t) + namelen;
  alloc_fileh();
  fh()->usize   = filelength(fileno(fp));
  fh()->namelen = uchar(namelen);
  strcpy(fh()->name(), filename);

  HEAPCHECK;
  GetFileTime(HANDLE(_get_osfhandle(fileno(fp))),
                                     &fh()->ctime, &fh()->atime, &fh()->mtime);

  // compress the file
  read_size = 0;
  int saved_zipped_size = zipped_size;
  HEAPCHECK;
  int code = fh()->usize ? zip_deflate(file_read, mem_write) : PK_OK;
  if ( read_size == -1 )
  {
    fclose(fp);
    return ERR_CANCEL;
  }
  switch ( code )
  {
    case PK_MEM:
      nomem("zip");
    case PK_WEOF:      // should not happen
      error("weof?");
    case PK_REOF:
      fclose(fp);
      return ERR_READ;
    default:
      error("ziperr %d", code);
    case PK_OK:
      fclose(fp);
      break;
  }
  fh()->csize = zipped_size - saved_zipped_size;
  h2->nfiles++;
  return ERR_OK;
}

//-----------------------------------------------------------------------
// initialize the compression routines, allocate the payload headers
// and put the "message" into the payload
void start_zip(char *message)
{
  // trim and align the message
  if ( message == NULL ) message = "";
  char *ptr = strchr(message, 0);
  while ( ptr > message && (ptr[-1] == '\r' || ptr[-1] == '\n') ) ptr--;
  *ptr = '\0';
  int msglen = strlen(message);
  if ( msglen != 0 )
  {
    msglen += 1 + 3;
    msglen &= ~3;
  }
  // allocate memory for the payload and initialize the payload headers
  header_size = sizeof(header_t) + msglen + sizeof(header2_t);
  memory_size = header_size;
  zipped_size = 0;
  free(memory);
  memory = (char*)malloc(memory_size+KEYSIZE);
  if ( memory == NULL ) nomem("start_zip");

  h = (header_t *)memory;
  h->magic    = MAGIC;
  h->msglen   = msglen;
  produce_iv(h->iv);
  // put the message after the first header
  if ( msglen ) strcpy(h->msg(), message);

  h2 = h->h2();
  h2->chk[0] = h->iv[0];
  h2->chk[1] = h->iv[1];
  h2->nfiles = 0;
}


